// $Id: TaskBarBase.cpp,v 1.95 2011/11/10 23:07:23 hpereira Exp $


/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

/*!
\file    TaskBar.cc
\brief   transparent frame to display TaskButton(s)
\author  Hugo Pereira
\version $Revision: 1.95 $
\date    $Date: 2011/11/10 23:07:23 $
*/

#include "Icons.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "Menu.h"
#include "QuestionDialog.h"
#include "RoundedRegion.h"
#include "SvgEngine.h"
#include "Singleton.h"
#include "TaskDialog.h"
#include "TaskBarBase.h"
#include "TaskBarConfiguration.h"
#include "TaskBarDialog.h"
#include "TaskListConfiguration.h"
#include "TaskListDialog.h"
#include "TaskParser.h"
#include "TaskToolTip.h"
#include "WinUtil.h"
#include "X11Util.h"
#include "XmlOptions.h"

#include <QtGui/QApplication>
#include <QtGui/QAction>
#include <QtGui/QDesktopWidget>
#include <QtGui/QMouseEvent>
#include <QtCore/QTextStream>

//_____________________________________
const QSize& TaskBarBase::emptySize( void )
{
    static QSize size( 350, 100 );
    return size;
}

//_____________________________________
TaskBarBase::TaskBarBase( QWidget *parent ):
    TransparentWidget( parent, Qt::FramelessWindowHint ),
    step_( 1 ),
    spacing_( 0 ),
    margin_( 0 ),
    nColumns_( 0 ),
    nLines_( 0 ),
    mouseButton_( Qt::NoButton ),
    enableMove_( false ),
    moved_( false ),
    needUpdate_( true ),
    contextMenu_( 0 ),
    tooltip_( new TaskToolTip( 0 ) ),
    action_( NONE ),
    useSvg_( true ),
    selectedButton_( TaskButton::null() )
{

    Debug::Throw( "TaskBarBase::TaskBarBase.\n" );

    /*
    disable all automatic background filling
    to optimize painting.
    */
    setAttribute( Qt::WA_OpaquePaintEvent, true );

    // actions
    _installActions();

    addAction( &reloadBackgroundAction() );

    // reset configuration
    connect( Singleton::get().application(), SIGNAL( configurationChanged() ), SLOT( updateConfiguration() ) );
    connect( qApp->desktop(), SIGNAL( resized( int ) ), SLOT( _desktopResized( void ) ) );

    _toolTip().setFocusProxy( this );
    setMouseTracking( true );

    //! initialize stored options
    storedOptions_.push_back( Options::Pair( "MIN_BUTTON_SIZE", "" ) );
    storedOptions_.push_back( Options::Pair( "MAX_BUTTON_SIZE", "" ) );
    storedOptions_.push_back( Options::Pair( "HIGHLIGHT_ON_ENTER", "" ) );
    storedOptions_.push_back( Options::Pair( "HIGHLIGHT_ON_CLICK", "" ) );
    storedOptions_.push_back( Options::Pair( "HIGHLIGHT_INTENSITY", "" ) );
    storedOptions_.push_back( Options::Pair( "HIGHLIGHT_STEPS", "" ) );
    storedOptions_.push_back( Options::Pair( "TRANSPARENCY_SHADOW_COLOR", "" ) );
    storedOptions_.push_back( Options::Pair( "TRANSPARENCY_SHADOW_OFFSET", "" ) );
    storedOptions_.push_back( Options::Pair( "TRANSPARENCY_FOREGROUND_COLOR", "" ) );

    connect( &SVG::SvgEngine::get(), SIGNAL( changed() ), SLOT( setBackgroundChanged() ) );

}

//_____________________________________
TaskBarBase::~TaskBarBase( void )
{ Debug::Throw( "TaskBarBase::~TaskBar.\n" ); }

//____________________________________________
void TaskBarBase::synchronizeAttributes( void )
{

    Debug::Throw( "TaskBarBase::synchronizeAttributes.\n" );

    TaskAttributes::List task_attributes;
    for( TaskButton::List::const_iterator iter = _buttons().begin(); iter != _buttons().end(); ++iter )
    {
        // retrieve task attributes
        TaskAttributes attributes( (*iter)->attributes() );

        // make sure creation time is valid
        if( !attributes.creation().isValid() )
        {
            if( attributes.modification().isValid() || attributes.lastCalled().isValid() ) attributes.setCreation( std::min(  attributes.modification(), attributes.lastCalled() ) );
            else attributes.setCreation( TimeStamp::now() );
        }

        // store in list
        task_attributes.push_back( attributes );

    }

    // update taskbar attributes
    attributes().setTaskAttributesList( task_attributes );

}
//____________________________________________
void TaskBarBase::setAttributes( const TaskBarAttributes& attributes, bool also_update )
{

    Debug::Throw( "TaskBarBase::setAttributes.\n" );

    // change flags
    for( ActionMap::const_iterator iter = flagActions().begin(); iter != flagActions().end(); ++iter )
    { iter->first->setChecked( attributes.flag( iter->second ) ); }

    // apply flags
    _toggleStaysOnTop( attributes.flag( TaskBarAttributes::STAYS_ON_TOP ) );
    _toggleHideFromTaskbar( attributes.flag( TaskBarAttributes::HIDE_FROM_TASKBAR ) );
    _toggleSticky( attributes.flag( TaskBarAttributes::STICKY ) );
    _toggleRoundCorners( attributes.flag( TaskBarAttributes::ROUND_CORNERS ) );

    // AUTO_HIDE is enabled only if layout match with position.
    flagAction( TaskBarAttributes::AUTO_HIDE ).setEnabled(
        ( (attributes.positionMode()&TaskBarAttributes::TOP_POSITION)    && attributes.layout() ==  TaskBarAttributes::HORIZONTAL_TOP ) ||
        ( (attributes.positionMode()&TaskBarAttributes::BOTTOM_POSITION) && attributes.layout() ==  TaskBarAttributes::HORIZONTAL_BOTTOM ) ||
        ( (attributes.positionMode()&TaskBarAttributes::LEFT_POSITION)   && attributes.layout() ==  TaskBarAttributes::VERTICAL_LEFT ) ||
        ( (attributes.positionMode()&TaskBarAttributes::RIGHT_POSITION) && attributes.layout() ==  TaskBarAttributes::VERTICAL_RIGHT ) );

    // see if buttons need update
    if( _optionsChanged() || TaskBarBase::attributes().needTaskUpdate( attributes ) )
    { _updateTasks( attributes ); }

    // store new attributes
    attributes_ = attributes;

    if( also_update )
    {

        // reset offset
        _setOffset( QPoint() );

        // resize from buttons
        _resize();
        _updatePosition();

        // move from attributes
        QtUtil::moveWidget( this, TaskBarBase::attributes().position() );
        setBackgroundChanged( true );
        _setNeedUpdate( true );

        // clear tooltip cache
        _toolTip().clearCache();

        update();
    }

}

//___________________________________________________
void TaskBarBase::updateConfiguration( void )
{

    Debug::Throw( "TaskBarBase::updateConfiguration.\n" );

    // opacity
    setWindowOpacity( XmlOptions::get().get<double>( "TRANSPARENCY_OPACITY" )/255 );

    _setStep( XmlOptions::get().get<unsigned int>( "GRID" ) );
    _setMargin( XmlOptions::get().get<unsigned int>( "MARGIN" ) );
    _setSpacing( XmlOptions::get().get<unsigned int>( "SPACING" ) );

    // tooltip
    _toolTip().setDelay( XmlOptions::get().get<unsigned int>("TOOLTIP_DELAY") );
    toggleToolTipAction().setChecked( XmlOptions::get().get<bool>("SHOW_TOOLTIP") );

    highlightOnClickAction().setChecked( XmlOptions::get().get<bool>("HIGHLIGHT_ON_CLICK" ) );
    highlightOnEnterAction().setChecked( XmlOptions::get().get<bool>("HIGHLIGHT_ON_ENTER" ) );
    noHighlightAction().setChecked( !(highlightOnClickAction().isChecked() || highlightOnEnterAction().isChecked() ) );

    // svg update
    _setUseSvg( XmlOptions::get().get<bool>( "USE_SVG" ) );

    // reload svg
    if( _useSvg() ) SVG::SvgEngine::get().reload();

    // reset attributes
    _resetAttributes();

}

//______________________________________________
void TaskBarBase::setMask( QRect mask )
{ setMask( mask, RoundedRegion::ALL ); }

//______________________________________________
void TaskBarBase::setMask( QRect mask, unsigned int corners )
{
    if(  attributes().flag( TaskBarAttributes::ROUND_CORNERS ) ) TRANSPARENCY::TransparentWidget::setMask( RoundedRegion( mask, corners ) );
    else {

        // for some reason, setMask does not work unless
        // one remove at least one pixmap from the rect
        // to make it a non rectangular region
        QRegion region( mask );
        region -= QRegion( mask.topLeft().x(), mask.topLeft().y(), 1, 1 );
        TRANSPARENCY::TransparentWidget::setMask( region );

    }
}

//____________________________________________
void TaskBarBase::_addButton( const bool& is_separator )
{
    Debug::Throw( "TaskBarBase::_addButton.\n" );
    TaskAttributes attributes;

    if( is_separator ) attributes.setFlag( TaskAttributes::IS_SEPARATOR, true );
    else {

        TaskDialog dialog( this );
        if( !dialog.centerOnPointer().exec() ) return;
        attributes = dialog.attributes();
        attributes.setModification();

    }

    // create TaskButton and check validity
    TaskButton::Pointer button( new TaskButton( TaskBarBase::attributes(), attributes ) );
    if( !button->isValid() )
    {
        InformationDialog( this, "Button is not valid. <add button> canceled." ).centerOnPointer().exec();
        button->deleteLater();
        return;
    }

    // update task list
    // check selected button

    TaskButton::Pointer selected( _selectedButton() );
    if( !selected ) _buttons().push_back( button );
    else {

        // try get current button and associated iterator
        TaskButton::List::iterator selectedIter = std::find_if( _buttons().begin(), _buttons().end(), TaskButton::SameAttributesFTor( selected->attributes() ) );

        // advance one, so that the new button is inserted "after" and not before the selected buttons
        if( selectedIter != _buttons().end() ) ++selectedIter;

        // insert button
        _buttons().insert( selectedIter, button );
    }


    // resize and save buttons
    _saveButtons();
    _resetAttributes();

}

//____________________________________________
void TaskBarBase::_removeButton( void )
{
    Debug::Throw( "TaskBarBase::Remove.\n" );

    if( !_selectedButton() )
    {
        InformationDialog( this, "no button selected. <Remove> canceled" ).setWindowTitle( "Remove Button - taskbar" ).centerOnPointer().exec();
        return;
    }

    // store selected button
    TaskAttributes attributes( _selectedButton()->attributes() );

    // confirm
    QString buffer;
    QTextStream what( &buffer );
    if( attributes.flag( TaskAttributes::IS_SEPARATOR ) ) { what << "Remove Separator ?"; }
    else { what << "Remove button " <<  attributes.name() << " ?"; }

    if( !QuestionDialog( this, buffer ).setWindowTitle( "Remove button - taskbar").centerOnPointer().exec() ) return;

    // remove task and save
    TaskButton::List::iterator iter;
    while( ( iter = find_if( _buttons().begin(), _buttons().end(), TaskButton::SameAttributesFTor( attributes ) ) ) != _buttons().end() )
    {
        (*iter)->deleteLater();
        _buttons().erase( iter );
    }

    // resize and save
    _saveButtons();
    _resetAttributes();

}

//____________________________________________
void TaskBarBase::_editButton( void )
{

    Debug::Throw( "TaskBarBase::Edit.\n" );
    if( !_selectedButton() )
    {
        InformationDialog( this, "no button selected. <Edit> canceled" ).centerOnPointer().exec();
        return;
    }

    // store selectedButton_ in case it changes while filling the dialog
    TaskButton::Pointer button = _selectedButton();

    if( button->attributes().flag( TaskAttributes::IS_SEPARATOR ) )
    {
        InformationDialog( this, "separators cannot be edited" ).centerOnPointer().exec();
        return;
    }

    // create dialog
    TaskDialog dialog( this );
    if( !dialog.setAttributes( button->attributes() ).centerOnPointer().exec() ) return;

    // retrieve attributes and update button (if needed)
    TaskAttributes attributes = dialog.attributes();
    if( attributes == button->attributes() ) return;

    // create new button
    attributes.setModification();
    TaskButton new_button( TaskBarBase::attributes(), attributes );
    if( !new_button.isValid() )
    {
        InformationDialog( this, "New attributes are not valid. <Edit> canceled." ).centerOnPointer().exec();
        return;
    }

    // update button
    button->setAttributes( TaskBarBase::attributes(), attributes );
    _setNeedUpdate(true);
    _saveButtons();
    _resetAttributes();

}

//____________________________________________
void TaskBarBase::_editAttributes( void )
{
    Debug::Throw( "TaskBarBase::_editAttributes.\n" );
    TaskBarDialog dialog( this, attributes() );
    if( !dialog.exec() ) return;

    TaskBarAttributes attributes( dialog.taskBarConfiguration().attributes() );
    if( attributes == TaskBarBase::attributes() )
    { Debug::Throw(0) << "TaskBarBase::_editAttributes - Attributes are unchanged." << endl; }

    emit attributesChanged( attributes );
    return;

}

//____________________________________________
void TaskBarBase::_editTaskList( void )
{
    Debug::Throw( "TaskBarBase::_editTaskList.\n" );
    TaskListDialog dialog( this, attributes().taskAttributesList() );
    if( !dialog.exec() ) return;

    TaskAttributes::List tasklist( dialog.taskListConfiguration().taskAttributesList() );
    if( tasklist == attributes().taskAttributesList() )
    { Debug::Throw(0) << "TaskBarBase::_editTaskList - Attributes are unchanged." << endl; }

    // update task list and reset
    attributes().setTaskAttributesList( tasklist );
    _resetAttributes();

    emit taskListChanged();

    return;
}

//___________________________________________________
void TaskBarBase::_setEnableMove( const bool& value )
{
    Debug::Throw( "TaskBarBase::_setEnableMove.\n" );
    if( enableMove_ == value ) return;

    enableMove_ = value;
    if( _moveEnabled() )
    {

        _setCursor( _mouseButton() );
        if( _selectedButton() ) { _leaveButton( _selectedButton() ); }
        _setAction( MINIMUM );

    } else {

        _setCursor( Qt::NoButton );
        _setAction( ENTER );

    }

    update();

}

//________________________________________________
void TaskBarBase::_toggleFlagAction( QAction* action )
{
    Debug::Throw( "TaskBarBase::_toggleFlagAction.\n" );
    ActionMap::const_iterator iter( flagActions().find( action ) );
    assert( iter != flagActions().end() );

    // check if state has changed
    if( action->isChecked() == attributes().flag( iter->second ) )
    {
        Debug::Throw() << "TaskBarBase::_toggleFlagAction - flag: " << iter->second << " unchanged" << endl;
        return;
    }

    // modify attributes accordingly
    TaskBarAttributes attributes( TaskBarBase::attributes() );
    attributes.setFlag( iter->second, action->isChecked() );
    emit attributesChanged( attributes );
    return;

}

//______________________________________________
void TaskBarBase::_toggleHighlight( QAction* action )
{

    Debug::Throw() << "TaskBarBase::_toggleHighlight" << endl;
    if( !action->isChecked() )
    { return; }

    XmlOptions::get().set<bool>( "HIGHLIGHT_ON_CLICK", highlightOnClickAction().isChecked() );
    XmlOptions::get().set<bool>( "HIGHLIGHT_ON_ENTER", highlightOnEnterAction().isChecked() );

    // loop over buttons
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); ++iter )
    {

        // skip separators of buttons that already have highlighted pixmaps
        if( (*iter)->attributes().flag( TaskAttributes::IS_SEPARATOR ) || (*iter)->hasHighlightedPixmaps() ) continue;

        // reset button attributes
        // to create/clear highlighted pixmaps
        (*iter)->setAttributes( attributes(), (*iter)->attributes() );

    }

}

//___________________________________________________
void TaskBarBase::_toggleToolTip( bool state )
{

    Debug::Throw( "TaskBarBase::_toggleToolTip.\n" );
    XmlOptions::get().set<bool>( "SHOW_TOOLTIP", state );

}

//__________________________________________________________
void TaskBarBase::_desktopResized( void )
{
    Debug::Throw( 0, "TaskbarBase::_destktopResized.\n" );

    if( attributes_.positionMode() == TaskBarAttributes::ANY_POSITION ) return;
    _updatePosition();

    // move from attributes
    QtUtil::moveWidget( this, TaskBarBase::attributes().position() );
    setBackgroundChanged( true );
    _setNeedUpdate( true );

    // clear tooltip cache
    _toolTip().clearCache();

    update();

}

//_____________________________________________________
#if defined(Q_WS_X11)
bool TaskBarBase::x11Event( XEvent* event )
{

    Debug::Throw( "TaskBarBase::x11Event.\n" );

    // cast event to a property event
    XPropertyEvent *property_event = (XPropertyEvent*) event;
    if(
        property_event &&
        property_event->window == winId() &&
        property_event->atom == X11Util::get().findAtom( X11Util::_NET_WM_STATE ) )
    {

        flagAction( TaskBarAttributes::STICKY ).setChecked(
            ( X11Util::get().isSupported( X11Util::_NET_WM_STATE_STICKY ) && X11Util::get().hasProperty( *this, X11Util::_NET_WM_STATE_STICKY ) ) ||
            ( X11Util::get().isSupported( X11Util::_NET_WM_DESKTOP ) && X11Util::get().cardinal( *this, X11Util::_NET_WM_DESKTOP ) == X11Util::ALL_DESKTOPS ) );
        flagAction( TaskBarAttributes::HIDE_FROM_TASKBAR ).setChecked( X11Util::get().hasProperty( *this, X11Util::_NET_WM_STATE_SKIP_TASKBAR ) );
        Debug::Throw( "TaskBarBase::x11Event. done.\n" );

        return true;

    } else return false;

}
#endif

//______________________________________________
void TaskBarBase::timerEvent( QTimerEvent* event )
{

    Debug::Throw( "TaskBarBase::timerEvent.\n" );

    if( event->timerId() == _updateTimer().timerId() )
    {
        update();
        _updateTimer().stop();
    } else if( event->timerId() == _enableMoveTimer().timerId() ) {

        _enableMoveTimer().stop();
        _setEnableMove();

        if( X11Util::get().moveWidget( *this, QCursor::pos() ) ) {

            _setMoved( true );
            _setMouseButton( Qt::NoButton );
            attributes().setPositionMode( TaskBarAttributes::ANY_POSITION );

        }


    } else if( event->timerId() == _geometryTimer().timerId() ) {
        _updateGeometry( _action() );
        _geometryTimer().stop();
        update();
    }

    else return TransparentWidget::timerEvent( event );

}

//______________________________________________
void TaskBarBase::resizeEvent( QResizeEvent* event )
{
    if( attributes().flag( TaskBarAttributes::ROUND_CORNERS ) ) _roundCorners();

    if( _blurEnabled() ) {

        static const int margin = 10;
        _updateBlurRegion( rect().translated( mapTo( window(), QPoint(0,0) ) ).adjusted( margin, margin, -margin, -margin ) );

    }

    TransparentWidget::resizeEvent( event );

}

//___________________________________________________
void TaskBarBase::enterEvent( QEvent* )
{

    Debug::Throw() << "TaskBarBase::enterEvent - button: " << _mouseButton() << " action: " << _action() << endl;

    // activate
    activateWindow();

    // emit signal
    emit taskBarEntered( attributes() );

    // reset cursor
    _setCursor( _mouseButton() );

    // try get button under pointer
    _setSelectedButton( _button( mapFromGlobal( QCursor::pos() ) ) );

    // if no button pressed zoom button
    if( _mouseButton() == Qt::NoButton && _selectedButton() && _action() == NONE )
    {
        _enterButton( _selectedButton() );
        update();

        // show tooltip
        if( toggleToolTipAction().isChecked() &&
            !_selectedButton()->attributes().flag( TaskAttributes::IS_SEPARATOR ) )
        {
            _toolTip().setAttributes( _selectedButton()->attributes() );
            _moveToolTip();
            _toolTip().show();
        }

    }

    // show
    if( !_moveEnabled() ) { _setAction( ENTER ); }
    update();

}

//___________________________________________________
void TaskBarBase::leaveEvent( QEvent* )
{

    Debug::Throw( "TaskBarBase::leaveEvent.\n" );

    // reset cursor
    _setCursor( Qt::NoButton );

    // disable move
    if( _mouseButton() == Qt::NoButton ) _setEnableMove( false );

    // unzoom _selectedButton and hide tooltip
    if( _selectedButton() ) _leaveButton( _selectedButton() );
    _setAction( LEAVE );
    update();

    // stop tooltip timers
    _toolTip().hide( true );

}

//___________________________________________________
void TaskBarBase::mousePressEvent( QMouseEvent *event )
{

    Debug::Throw( "TaskBarBase::mousePressEvent.\n" );

    // retrieve button
    TaskButton::Pointer current = _button( event->pos() );

    // hide tooltip
    _toolTip().hide( true );

    // check if button has changed
    if( current && current == _selectedButton() && event->button() == Qt::LeftButton )
    {
        _pressButton( _selectedButton() );
        update();
    }

    // update selection and store button state
    _setSelectedButton( current );
    _setMouseButton( event->button() );
    _setCursor( _mouseButton() );

    // store position if left button (for moving)
    if( _mouseButton() == Qt::LeftButton ) {

        _setMoved( false );
        if( !( attributes().flag( TaskBarAttributes::LOCK_POSITION ) ) )
        {
            _setClickPosition( event->pos() + QPoint(geometry().topLeft() - frameGeometry().topLeft()) );
            _enableMoveTimer().start( 200, this );
        }

        return;
    }

    // mid button change cursor
    if( _mouseButton() == Qt::MidButton ) {
        return;
    }

    // ask for menu in case of right button
    if( _mouseButton() == Qt::RightButton )
    {

        /*
        when raising the menu, the pointer leaves the window
        and it is likely that we don't get the release event.
        as such we force the button state to Qt::NoButton now
        */
        _setMouseButton( Qt::NoButton );
        return QWidget::mousePressEvent( event );

    }

}

//___________________________________________________
void TaskBarBase::mouseReleaseEvent( QMouseEvent *event )
{
    Debug::Throw( "TaskBarBase::mouseReleaseEvent.\n" );

    // retrieve button
    TaskButton::Pointer current = _button( event->pos() );

    // check if button has changed
    if( current && current == _selectedButton() && event->button() == Qt::LeftButton )
    {
        _releaseButton( _selectedButton() );
        update();
    }

    // left button, run application if not moved; disable move
    if(  _mouseButton() == Qt::LeftButton )
    {

        if( current && !_moved() && !current->attributes().exec() )
        {
            QString buffer;
            QTextStream( &buffer )
                << "Could not execute command " << endl
                << current->attributes().command();
            InformationDialog( this, buffer ).centerOnPointer().exec();
        }

        _enableMoveTimer().stop();

        // left button disable movement
        if( _moveEnabled() ) _setEnableMove( false );
    }

    // mid button end drag/drop
    if( _mouseButton() == Qt::MidButton )
    { _saveButtons(); }

    // update selection and store button state
    _setCursor( Qt::NoButton );
    _setMoved( false );
    _setSelectedButton( current );
    _setMouseButton( Qt::NoButton );

}

//___________________________________________________
void TaskBarBase::mouseMoveEvent( QMouseEvent *event )
{

    Debug::Throw( "TaskBarBase::mouseMoveEvent.\n" );

    // left button, move the window
    if( _mouseButton() == Qt::LeftButton && !attributes().flag( TaskBarAttributes::LOCK_POSITION ) && _moveEnabled() )
    {
        QPoint true_point(event->globalPos() - _clickPosition() );
        QPoint point(
            _step()*int(true_point.x()/_step()),
            _step()*int(true_point.y()/_step())
            );

        // store position and disable auto_hide
        attributes().setPosition( point );
        attributes().setPositionMode( TaskBarAttributes::ANY_POSITION );
        move( point );

        // change move flag
        _setMoved( true );

        return;

    }

    // store mouse position
    _setMousePosition( event->pos() );

    // retrieve button
    TaskButton::Pointer current = _button( event->pos() );

    // middle button, drag
    if(
        (!attributes().flag( TaskBarAttributes::LOCK_BUTTONS )) &&
        _mouseButton() == Qt::MidButton )
        _swap( _selectedButton(), current );

    // unselect old button
    bool need_zoom( false );
    if( _selectedButton() && _selectedButton() != current )
    {
        _leaveButton( _selectedButton() );
        need_zoom = true;

        // hide tooltip
        if( _mouseButton() == Qt::NoButton && (!current || current->attributes().flag( TaskAttributes::IS_SEPARATOR ) ) )
        { _toolTip().hide( false ); }

    }

    // select new button
    if( current && _selectedButton() != current && _action() == NONE )
    {
        _setSelectedButton( current );
        _enterButton( _selectedButton() );
        need_zoom = true;

        // show tooltip
        if(
            _mouseButton() == Qt::NoButton &&
            toggleToolTipAction().isChecked() &&
            !current->attributes().flag( TaskAttributes::IS_SEPARATOR ) )
        {
            _toolTip().setAttributes( current->attributes() );
            _moveToolTip();
            _toolTip().show();
        }

    }

    if( need_zoom )  update();

}

//________________________________________________
void TaskBarBase::contextMenuEvent( QContextMenuEvent* event )
{

    Debug::Throw( "TaskBarBase::contextMenuEvent.\n" );

    // update actions
    editButtonAction().setEnabled( _selectedButton() && !_selectedButton()->attributes().flag( TaskAttributes::IS_SEPARATOR ) );
    removeButtonAction().setEnabled( _selectedButton() );
    if( _selectedButton() )
    { removeButtonAction().setText( _selectedButton()->attributes().flag( TaskAttributes::IS_SEPARATOR ) ? "&Remove separator":"&Remove button" ); }

    _contextMenu().exec( event->globalPos() );

}

//___________________________________________________
void TaskBarBase::_paint( QPaintDevice& device, const QRect& rect )
{

    Debug::Throw( "TaskBarBase::_paint.\n" );

    if( _backgroundChanged() )
    {
        _updateBackgroundPixmap();
        _setNeedUpdate( true );
    }

    // draw either empty taskbar of buttons if any
    if( _buttons().empty() ) { _empty( device, rect ); }
    else if( _action() == ENTER ) {

        _drawButtons( device, rect );
        _geometryTimer().start( attributes().zoomDelay(), this );

    } else if( _drawButtons( device, rect ) ) {

        _updateTimer().start( attributes().zoomDelay(), this );

    } else if( _action() == LEAVE || _action() == MINIMUM ) {

        _drawButtons( device, rect );
        _geometryTimer().start( attributes().zoomDelay(), this );

    }

}

//________________________________________________
Menu* TaskBarBase::_createContextMenu( void )
{

    Debug::Throw( "TaskBarBase::_createDefaultContextMenu.\n" );
    return new Menu( this );

}

//___________________________________________________
void TaskBarBase::_saveButtons( void )
{

    Debug::Throw( "TaskBarBase::_saveButtons.\n" );
    attributes_.clear();
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); ++iter )
    { attributes().addTaskAttributes( (*iter)->attributes() ); }

    emit modified();

    Debug::Throw( "TaskBarBase::_saveButtons - done.\n" );

    return;

}


//___________________________________________________
void TaskBarBase::_clearButtons( void )
{

    Debug::Throw( "TaskBarBase::_clearButtons.\n" );
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); ++iter )
    { (*iter)->deleteLater(); }

    _buttons().clear();

}

//___________________________________________________
void TaskBarBase::_setDimensions( void )
{
    const TaskBarAttributes::Layout &layout( attributes().layout() );
    if( layout & TaskBarAttributes::HORIZONTAL )
    {

        nLines_ = 1;
        nColumns_ = _buttons().size();

    } else if( layout & TaskBarAttributes::VERTICAL ) {

        nLines_ = _buttons().size();
        nColumns_ = 1;

    } else if( layout & TaskBarAttributes::BOX ) {

        nColumns_ = attributes().numColumns();
        nLines_ = _buttons().size()/nColumns_;
        while( nLines_*nColumns_ < _buttons().size() ) nLines_ ++;

    } else assert(0);

}

//_______________________________________________________________________________
void TaskBarBase::_enterButton( TaskButton::Pointer button )
{
    Debug::Throw( "TaskBarBase::_enterButton.\n" );
    if( highlightOnEnterAction().isChecked() || ( _mouseButton() == Qt::LeftButton && highlightOnClickAction().isChecked() ) )
    { button->highlight( true ); }

    button->setDirection( TaskButton::BIGGER );
}

//_______________________________________________________________________________
void TaskBarBase::_leaveButton( TaskButton::Pointer button )
{
    Debug::Throw( "TaskBarBase::_leaveButton.\n" );
    button->setDirection( TaskButton::SMALLER );
    if( highlightOnEnterAction().isChecked() || highlightOnClickAction().isChecked() ) button->highlight( false );
}

//_______________________________________________________________________________
void TaskBarBase::_pressButton( TaskButton::Pointer button )
{
    Debug::Throw( "TaskBarBase::_pressButton.\n" );
    if( highlightOnClickAction().isChecked() ) {
        button->highlight( true );
        button->setDirection( TaskButton::BIGGER );
    } else {
        button->setDirection( TaskButton::SMALLER );
    }

}

//_______________________________________________________________________________
void TaskBarBase::_releaseButton( TaskButton::Pointer button )
{ button->setDirection( TaskButton::BIGGER ); }

//_______________________________________________________________________________
void TaskBarBase::_swap( TaskButton::Pointer first, TaskButton::Pointer second )
{
    // check buttons
    if( !( first && second && first != second ) ) return;

    // retrieve attributes
    TaskAttributes source( first->attributes() );
    TaskAttributes dest( second->attributes() );

    // check no button is separator
    if( source.flag( TaskAttributes::IS_SEPARATOR ) ) return;

    // swap attributes
    first->setAttributes( attributes(), dest );
    second->setAttributes( attributes(), source );

}

//________________________________________________
void TaskBarBase::_toggleStaysOnTop( bool state )
{
    Debug::Throw( "TaskBarBase::_toggleStaysOnTop.\n" );

    // check that widget is top level
    if( parentWidget() ) return;

    #if defined(Q_WS_X11)

    // change property
    X11Util::get().changeProperty( *this, X11Util::_NET_WM_STATE_STAYS_ON_TOP, state );
    X11Util::get().changeProperty( *this, X11Util::_NET_WM_STATE_ABOVE, state );

    #else

    if( state == (windowFlags() | Qt::WindowStaysOnTopHint) ) return;

    bool visible( !isHidden() );
    if( state ) setWindowFlags( windowFlags() | Qt::WindowStaysOnTopHint );
    else setWindowFlags( windowFlags() & ~Qt::WindowStaysOnTopHint );
    if( visible ) show();

    #endif

}

//________________________________________________
void TaskBarBase::_toggleHideFromTaskbar( bool state )
{

    Debug::Throw() << "TaskBarBase::_toggleHideFromTaskbar - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if defined(Q_WS_X11)
    X11Util::get().changeProperty( *this, X11Util::_NET_WM_STATE_SKIP_TASKBAR, state );
    X11Util::get().changeProperty( *this, X11Util::_NET_WM_STATE_SKIP_PAGER, state );
    #endif

    #if defined(Q_WS_WIN)
    WinUtil( this ).toggleHideFromTaskbar( state );
    #endif


    return;

}

//________________________________________________
void TaskBarBase::_toggleSticky( bool state )
{

    Debug::Throw() << "TaskBarBase::_toggleSticky - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if defined(Q_WS_X11)
    if( X11Util::get().isSupported( X11Util::_NET_WM_STATE_STICKY ) )
    {

        // change property depending on state
        X11Util::get().changeProperty( *this, X11Util::_NET_WM_STATE_STICKY, 1);

    } else if( X11Util::get().isSupported( X11Util::_NET_WM_DESKTOP ) ) {

        unsigned long desktop = X11Util::get().cardinal( QX11Info::appRootWindow(), X11Util::_NET_CURRENT_DESKTOP );
        X11Util::get().changeCardinal( *this, X11Util::_NET_WM_DESKTOP, state ? X11Util::ALL_DESKTOPS:desktop );

    }

    #endif

    return;

}

//________________________________________________
void TaskBarBase::_toggleRoundCorners( bool state )
{
    Debug::Throw( "TaskBarBase::_toggleRoundCorners.\n" );
    if( state ) _roundCorners();
    else clearMask();
}

//___________________________________________________
void TaskBarBase::_empty( QPaintDevice& device, const QRect& clip )
{

    Debug::Throw( "TaskBarBase::_Empty.\n" );

    // create painter on foreground pixmap
    QPainter painter( &device );
    painter.setClipRect( clip );

    painter.setRenderHints(QPainter::SmoothPixmapTransform);
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver);

    // draw background pixmap into entire window
    painter.drawPixmap( 0, 0, _backgroundPixmap() );

    QRect rect( TaskBarBase::rect() );
    int alignment = Qt::AlignCenter;
    QString text( "the taskbar is empty\nright click to add button" );

    // update font
    QFont font( painter.font() );
    font.setPointSize( 20 );
    font.setBold( true );
    painter.setFont( font );

    QColor shadow_color( XmlOptions::get().get<QString>( "TRANSPARENCY_SHADOW_COLOR" ) );
    int shadow_offset( XmlOptions::get().get<int>( "TRANSPARENCY_SHADOW_OFFSET" ) );
    if( shadow_offset != 0 && shadow_color.isValid() )
    {
        painter.setPen( shadow_color );
        painter.drawText(
            rect.left() + shadow_offset,
            rect.top() + shadow_offset,
            rect.width(),
            rect.height(),
            alignment,
            text );
    }

    QColor foreground_color( XmlOptions::get().get<QString>( "TRANSPARENCY_FOREGROUND_COLOR" ) );
    if( !foreground_color.isValid() ) foreground_color = palette().color( foregroundRole() );
    painter.setPen( foreground_color );
    painter.drawText(
        rect.left(),
        rect.top(),
        rect.width(),
        rect.height(),
        alignment,
        text );
    painter.end();

}

//___________________________________________________
void TaskBarBase::_updatePosition( void )
{

    Debug::Throw( "TaskBarBase::_updatePosition.\n" );
    if( attributes_.positionMode() == TaskBarAttributes::ANY_POSITION ) return;

    // retrieve desktop position and origin
    QRect screen = qApp->desktop()->screenGeometry( qApp->desktop()->screenNumber( this ) );

    // only set desktop size
    QSize desktop_size( screen.size() );
    QPoint point( screen.topLeft() );

    // this widget size
    const QSize& size( TaskBarBase::size() );

    switch( attributes_.positionMode() )
    {
        case TaskBarAttributes::TOP_LEFT_POSITION:
        break;

        case TaskBarAttributes::TOP_CENTER_POSITION:
        point.rx()+= ( desktop_size.width() - size.width() )/2;
        break;

        case TaskBarAttributes::TOP_RIGHT_POSITION:
        point.rx()+= ( desktop_size.width() - size.width() );
        break;

        case TaskBarAttributes::BOTTOM_LEFT_POSITION:
        point.ry() += ( desktop_size.height() - size.height() );
        break;

        case TaskBarAttributes::BOTTOM_CENTER_POSITION:
        point += QPoint( ( desktop_size.width() - size.width() )/2, desktop_size.height() - size.height() );
        break;

        case TaskBarAttributes::BOTTOM_RIGHT_POSITION:
        point += QPoint( desktop_size.width() - size.width(), desktop_size.height() - size.height() );
        break;

        case TaskBarAttributes::LEFT_CENTER_POSITION:
        point.ry() += ( desktop_size.height() - size.height() )/2;
        break;

        case TaskBarAttributes::RIGHT_CENTER_POSITION:
        point += QPoint( desktop_size.width() - size.width()-1, ( desktop_size.height() - size.height() )/2 );
        break;

        default:
        break;
    }

    attributes_.setPosition( point );

    return;
}

//___________________________________________________
void TaskBarBase::_moveToolTip( void )
{
    Debug::Throw( "TaskBarBase::_moveToolTip.\n" );

    // desktop geometry
    QDesktopWidget* desktop( qApp->desktop() );
    QRect desktop_geometry( desktop->screenGeometry( desktop->screenNumber( this ) ) );

    // tooltip size
    const QSize& tooltip_size( _toolTip().adjustedSize() );

    // display geometry
    QRect display_geometry( frameGeometry() );

    // button geometry
    QRect button_geometry( _selectedButton()->rect() );

    // final position
    QPoint position;

    // decide Widget position
    const TaskBarAttributes::Layout& layout( attributes().layout() );
    if( layout & (TaskBarAttributes::BOX | TaskBarAttributes::HORIZONTAL) )
    {

        // horizontal center must match the selected button center
        int x = int(
            mapToGlobal( button_geometry.bottomLeft() ).x() +
            0.5*(button_geometry.width() - tooltip_size.width() ) );

        if( x < 0 ) x = 0;
        if( x+tooltip_size.width()>desktop_geometry.width() )
            x = desktop_geometry.width()-tooltip_size.width();

        // try put above
        int y =
            display_geometry.topLeft().y() -
            tooltip_size.height() - _margin();

        // put bellow if failed
        if( y < 0 ) y = display_geometry.bottomLeft().y() + 2*_margin();

        // save position
        position = QPoint( x,y );

    } else if( layout & TaskBarAttributes::VERTICAL )
    {

        // vertical center must match the selected button center
        int y = int(
            mapToGlobal( button_geometry.topLeft() ).y() +
            0.5*(button_geometry.height() - tooltip_size.height() ) );

        if( y < 0 ) y = 0;
        if( y+tooltip_size.height()>desktop_geometry.height() )
            y = desktop_geometry.height()-tooltip_size.height();

        // try put left
        int x =
            display_geometry.topLeft().x() -
            tooltip_size.width() - _spacing();

        // try put right if failed
        if( x<0 ) x = display_geometry.topRight().x() + _spacing();

        // save position
        position = QPoint( x,y );

    } else assert(0);

    _toolTip().setUpdatesEnabled( false );

    //_toolTip().move( position );
    _toolTip().setGeometry( QRect( position, tooltip_size ) );
    _toolTip().updateBackgroundAction().trigger();
    _toolTip().setUpdatesEnabled( true );

}

//____________________________________
void TaskBarBase::_reloadBackground( void )
{
    Debug::Throw( "TaskBarBase::_reloadBackground.\n" );
    if( _useSvg() ) SVG::SvgEngine::get().reload( true );
    TransparentWidget::_reloadBackground();
}

//____________________________________
TaskButton::Pointer TaskBarBase::_button( const QPoint& point )
{
    Debug::Throw( "TaskBarBase::_button.\n" );
    TaskButton::List::iterator iter( find_if( _buttons().begin(), _buttons().end(), TaskButton::ContainsFTor( point ) ) );
    return (iter == _buttons().end()) ? TaskButton::null() : *iter;
}

//____________________________________
void TaskBarBase::_setCursor( const Qt::MouseButton& state )
{

    if( state == Qt::LeftButton && _moveEnabled() )  setCursor( Qt::SizeAllCursor );
    else if( state == Qt::MidButton && !attributes().flag( TaskBarAttributes::LOCK_BUTTONS ) )
    {

        if( attributes().layout() & TaskBarAttributes::HORIZONTAL ) setCursor( Qt::SizeHorCursor );
        else if( attributes().layout() & TaskBarAttributes::VERTICAL ) setCursor( Qt::SizeVerCursor );
        else setCursor( Qt::SizeAllCursor );

    } else unsetCursor();

}

//____________________________________________________________
void TaskBarBase::_setAction( const TaskBarBase::Action& action )
{
    Debug::Throw() << "TaskBarBase::_setAction - action: " << action << endl;
    action_ = action;
}

//___________________________________________________
bool TaskBarBase::_updateGeometry( TaskBarBase::Action action )
{
    Debug::Throw( "TaskBarBase::_updateGeometry.\n" );

    if( !attributes().flag( TaskBarAttributes::AUTO_HIDE ) )
    {
        _setAction( NONE );
        return false;
    }

    if( action == NONE ) return false;

    // by default reset action to NONE
    _setAction( NONE );

    // window is to be minimized
    if( action == MINIMUM )
    {
        _setOffset( QPoint() );
        return false;
    }


    QRect mask( _minimumRect() );
    unsigned int corners(0);

    if( (attributes().positionMode()&TaskBarAttributes::BOTTOM_POSITION) && attributes().layout() == TaskBarAttributes::HORIZONTAL_BOTTOM )
    {

        // BOTTOM TASKBAR
        if(
            ( action == ENTER && _offset().y() <= 0 ) ||
            ( action == LEAVE && _offset().y() >= mask.height()-1 ) ) return false;
        _offset().ry() += ( action == ENTER ) ? -1:1;
        mask.setTop( std::min( mask.top() + _offset().y(), mask.bottom()-2 ) );
        corners = RoundedRegion::TOP_LEFT|RoundedRegion::TOP_RIGHT;

    } else if( ( attributes().positionMode()&TaskBarAttributes::TOP_POSITION ) && attributes().layout() ==  TaskBarAttributes::HORIZONTAL_TOP ) {

        // TOP TASKBAR
        if(
            ( action == ENTER && _offset().y() >= 0 ) ||
            ( action == LEAVE && _offset().y() + mask.height() <= 1 ) ) return false;

        _offset().ry() += ( action == ENTER ) ? 1:-1;
        mask.setBottom( std::max( mask.bottom() + _offset().y(), mask.top()+2 ) );
        corners = RoundedRegion::BOTTOM_LEFT|RoundedRegion::BOTTOM_RIGHT;

    } else if ( (attributes().positionMode()&TaskBarAttributes::LEFT_POSITION) && attributes().layout() ==  TaskBarAttributes::VERTICAL_LEFT ) {

        // LEFT POSITION
        if(
            ( action == ENTER && _offset().x() >= 0 ) ||
            ( action == LEAVE && _offset().x() + mask.width() <= 1 ) ) return false;
        _offset().rx() += ( action == ENTER ) ? 1:-1;
        mask.setRight( std::max( mask.right() + _offset().x(), mask.left()+2) );
        corners = RoundedRegion::TOP_RIGHT|RoundedRegion::BOTTOM_RIGHT;

    } else if( (attributes().positionMode()&TaskBarAttributes::RIGHT_POSITION) && attributes().layout() ==  TaskBarAttributes::VERTICAL_RIGHT ) {

        // RIGHT POSITION
        if(
            ( action == ENTER && _offset().x() <= 0 ) ||
            ( action == LEAVE && _offset().x() >= mask.width()-1 ) ) return false;

        _offset().rx() += ( action == ENTER ) ? -1:1;
        mask.setLeft( std::min( mask.left() + _offset().x(), mask.right()-2) );
        corners = RoundedRegion::TOP_LEFT|RoundedRegion::BOTTOM_LEFT;

    } else {

        return false;

    }

    // action is left to its current value
    // indicating that further updates are needed
    _setNeedUpdate( true );
    _setAction( action );
    setMask( mask, corners );

    return true;

}

//____________________________________
void TaskBarBase::_installActions( void )
{
    Debug::Throw( "TaskBarBase::_installAction" );

    addAction( addButtonAction_ = new QAction( IconEngine::get( ICONS::ADD ), "&Add Button", this ) );
    connect( addButtonAction_, SIGNAL( triggered() ), SLOT( _addButton() ) );

    addAction( addSeparatorAction_ = new QAction( IconEngine::get( ICONS::ADD ), "Add &Separator", this ) );
    connect( addSeparatorAction_, SIGNAL( triggered() ), SLOT( _addSeparator() ) );

    addAction( removeButtonAction_ = new QAction( IconEngine::get( ICONS::REMOVE ), "&Remove Button", this ) );
    connect( removeButtonAction_, SIGNAL( triggered() ), SLOT( _removeButton() ) );

    addAction( editButtonAction_ = new QAction( IconEngine::get( ICONS::CONFIGURE ), "&Edit Button", this ) );
    connect( editButtonAction_, SIGNAL( triggered() ), SLOT( _editButton() ) );

    addAction( editAttributesAction_ = new QAction( IconEngine::get( ICONS::CONFIGURE ), "&Edit Taskbar Attributes", this ) );
    connect( editAttributesAction_, SIGNAL( triggered() ), SLOT( _editAttributes() ) );

    addAction( editTaskListAction_ = new QAction( IconEngine::get( ICONS::CONFIGURE ), "&Edit Button List", this ) );
    connect( editTaskListAction_, SIGNAL( triggered() ), SLOT( _editTaskList() ) );

    addAction( toggleTooltipAction_ = new QAction( "&Show Tooltip", this ) );
    toggleTooltipAction_->setCheckable( true );
    connect( toggleTooltipAction_, SIGNAL( toggled( bool ) ), SLOT( _toggleToolTip( bool ) ) );

    QActionGroup* action_group = new QActionGroup( this );
    action_group->setExclusive( true );
    connect( action_group, SIGNAL( triggered( QAction* ) ), SLOT( _toggleHighlight( QAction* ) ) );

    addAction( noHighlightAction_ = new QAction( "&No Highlight", this ) );
    noHighlightAction_->setCheckable( true );
    action_group->addAction(noHighlightAction_ );

    addAction( highlightOnEnterAction_ = new QAction( "&Highlight on Enter", this ) );
    highlightOnEnterAction_->setCheckable( true );
    action_group->addAction(highlightOnEnterAction_ );

    addAction( highlightOnClickAction_ = new QAction( "&Highlight on Click", this ) );
    highlightOnClickAction_->setCheckable( true );
    action_group->addAction(highlightOnClickAction_ );

    // flag action group
    QActionGroup* group( new QActionGroup( this ) );
    group->setExclusive( false );

    // lock position
    QAction* action;
    addAction( action = new QAction( "&Lock Position", this ) );
    action->setToolTip( "Lock taskbar position" );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::LOCK_POSITION ) );

    // lock position
    addAction( action = new QAction( "&Auto Hide", this ) );
    action->setToolTip( "Auto hide taskbar" );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::AUTO_HIDE ) );

    // stays on top
    addAction( action = new QAction( "&Stays on Top", this ) );
    action->setToolTip( "Keep window on top of all others" );
    action->setCheckable( true );
    //action->setChecked( windowFlags() & Qt::WindowStaysOnTopHint );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::STAYS_ON_TOP ) );

    // sticky
    addAction( action = new QAction( "&Sticky", this ) );
    action->setToolTip( "Make window appear on all desktops" );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::STICKY ) );

    // hide from taskbar
    addAction( action = new QAction( "&Hide from Taskbar", this ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::HIDE_FROM_TASKBAR ) );

    // round corners
    addAction( action = new QAction( "&Round Corners", this ) );
    action->setToolTip( "Round window corners" );
    action->setCheckable( true );
    action->setChecked( false );
    action->setEnabled( windowFlags() & Qt::FramelessWindowHint );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::ROUND_CORNERS ) );

    // lock button position
    addAction( action = new QAction( "&Lock Button Position", this ) );
    action->setToolTip( "Lock buttons position" );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::LOCK_BUTTONS ) );

    // show button name
    addAction( action = new QAction( "&Show Button Names", this ) );
    action->setToolTip( "Show button names below icon" );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::SHOW_NAME ) );

    // expand separators
    addAction( action = new QAction( "&Expand Separators", this ) );
    action->setToolTip( "Expand separators when zooming" );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( std::make_pair( action, TaskBarAttributes::EXPAND_SEPARATORS ) );

    connect( group, SIGNAL( triggered( QAction* ) ), SLOT( _toggleFlagAction( QAction* ) ) );

}

//________________________________________________
void TaskBarBase::_roundCorners( void )
{ setMask( rect() ); }

//________________________________________________
bool TaskBarBase::_optionsChanged( void )
{

    Debug::Throw( "TaskBarBase::_optionsChanged.\n" );
    bool out( false );
    for( std::vector<Options::Pair>::iterator iter = storedOptions_.begin(); iter != storedOptions_.end(); ++iter )
    {
        Option current;
        if( XmlOptions::get().find( iter->first ) && (current = XmlOptions::get().option( iter->first )) == iter->second ) continue;

        Debug::Throw() << "TaskBarBase::_optionsChanged -"
            << " option: " << iter->first
            << " old value: " << iter->second.raw()
            << " new value: " << current.raw()
            << endl;
        iter->second = current;
        out = true;

    }

    return out;

}

//________________________________________________
void TaskBarBase::_updateTasks( const TaskBarAttributes& attributes )
{

    Debug::Throw( "TaskBarBase::_updateTasks.\n" );

    // reset
    _setSelectedButton( TaskButton::null() );
    _clearButtons();

    Debug::Throw( "TaskBarBase::_updateTasks - buttons cleared.\n" );
    for( TaskAttributes::List::const_iterator iter = attributes.taskAttributesList().begin(); iter != attributes.taskAttributesList().end(); ++iter )
    {

        // create button, check validity and add to list
        TaskButton::Pointer button( new TaskButton( attributes, *iter ) );
        if( button->isValid() ) _buttons().push_back( button );
        else button->deleteLater();

    }

    Debug::Throw( "TaskBarBase::_updateTasks - done.\n" );

}
